<?php

/*
 * Plugin: ultimania
 * ~~~~~~~~~~~~~~~~~
 * This plugin is made to replace the database Gerymania. Gerymania was sadly
 * the only global record database supporting the Stunts mode. This wouldn't be
 * the problem, if it would be fast and have just that basic features like
 * a ban system for cheaters (in stunt, cheating is very easy :/).
 * This is the reason, why this plugin have been coded. In the feature it'll
 * include:
 * 1. A ban system
 * 2. A webversion to browse the records
 * 3. A "report" function ingame and in the browser to report cheats
 * 		(those reported records will be checked by ultimania moderators)
 * 4. An export function to import the records from gerymania to ultimania.
 * 		(this isn't included in the plugin, theres a script which loads the known
 * 		UIds from the database and request the top rankings)
 * 
 * ----------------------------------------------------------------------------------
 * Author:		askuri
 * Version:		See below
 * Date:		07.12.2013
 * Copyright:	2013 Martin Weber
 * Hompeage:	askuri.de
 * Xaseco:		XAseco/1.14+
 * Game:		Trackmania Forever (TMF) only
 * ----------------------------------------------------------------------------------
 *
 * LICENSE: This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * ----------------------------------------------------------------------------------
 *
 * Dependencies: none
 * 
 * Manialink-IDs:
 * - ultimania_widget
 * - ultimania_window
 * 
 * 
 * Action-IDs:
 * - 5450101 = Open records window
 * - 5450102 = Close records window
 * - 5450103 = Report Map
 */

Aseco::registerEvent('onSync',						'ultimania_onSync');
Aseco::registerEvent('onShutdown',					'ultimania_onShutdown');
Aseco::registerEvent('onNewChallenge2',				'ultimania_onNewChallenge2');
Aseco::registerEvent('onNewChallenge2',				'ultimania_refreshInfo');
Aseco::registerEvent('onPlayerFinish',				'ultimania_onPlayerFinish');
Aseco::registerEvent('onPlayerConnect',				'ultimania_onPlayerConnect');
Aseco::registerEvent('onPlayerConnect',				'ultimania_widgetShow');
Aseco::registerEvent('onEndRace',					'ultimania_endRace');
Aseco::registerEvent('onEverySecond',				'ultimania_onEverySecond');
Aseco::registerEvent('onPlayerManialinkPageAnswer', 'ultimania_mlAnswer');

Aseco::addChatCommand('ultirankinfo', 'Shows informations about a record', false);

define ('ULTI_VERSION',		'0.2.0');
define ('ULTI_API_VERSION', '2');
define ('ULTI_API_INFO', 	'http://askuri.de/ultimania/'); // This URL is permanent. It will (ok, should) never change and defines the real home of the API

global $ulti_cfg, $ultimania_db, $ultimania_locals;
$ultimania_db = false;
$ultimania_locals = array();

function chat_ultirankinfo($aseco, $command) {
	global $ultimania_db;
	
	$rank = $command['params'] -= 1;
	
	if (!is_numeric($rank)) $aseco->client->query('ChatSendServerMessageToLogin', 'Usage: /ultirankinfo ranknumber  (e.g. /ultirankinfo 1)', $player->login);
	if (!isset($ultimania_db[$rank])) $aseco->client->query('ChatSendServerMessageToLogin', 'This record does not exists', $player->login);
	
	$header = 'Ultimania record information:';
	
	$time = date('M j Y G:i:s', $ultimania_db[$rank]['add_time']);
	
	$data = array();
	$data[] = array('Login:         $FFC'.$ultimania_db[$rank]['login']);
	$data[] = array('Nickname:  $FFC'.$ultimania_db[$rank]['nick']);
	$data[] = array('Driven:      $FFC'.$time.' (CET)');
	$data[] = array();
	$data[] = array('$h[ulti:admin_rec?uid='.$aseco->server->challenge->uid.'&login='.$ultimania_db[$command['params']]['login'].']Admin');
	
	display_manialink($command['author']->login, $header, array('Icons64x64_1', 'TrackInfo', -0.01), $data, array(0.9), 'OK');
}

function ultimania_mlAnswer($aseco, $answer) {
	switch ($answer[2]) {
		case 5450101:
			ultimania_windowShow($aseco, $answer[1]);
			break;
		case 5450102:
			ultimania_windowHide($aseco, $answer[1]);
			break;
		case 5450103:
			ultimania_reportMap($aseco, $answer[1]);
			break;
	}
}

function ultimania_endRace($aseco, $data) {
	ultimania_widgetHide($aseco);
	ultimania_windowHide($aseco, false);
}

function ultimania_refreshInfo() {
	global $ulti_cfg;
	
	$ulti_cfg['api_url'] 		= trim(file_get_contents(ULTI_API_INFO.'url.txt')).'TMF/'.ULTI_API_VERSION.'/index.php';
	$ulti_cfg['window_welcome'] = file_get_contents(ULTI_API_INFO.'tmf/description_window.txt');
}

// called @ onSync
function ultimania_onSync($aseco) {
	global $ulti_cfg;
	
	// Check for the right XAseco-Version
	$xaseco_min_version = '1.14';
	if ( defined('XASECO_VERSION') ) {
		if ( version_compare(XASECO_VERSION, $xaseco_min_version, '<') ) {
			trigger_error('[plugin.ultimania.php] Not supported XAseco version ('. XASECO_VERSION .')! Please update to min. version '. $xaseco_min_version .'!', E_USER_ERROR);
		}
	}
	else {
		trigger_error('[plugin.ultimania.php] Can not identify the System, "XASECO_VERSION" is unset! This plugin runs only with XAseco/'. $xaseco_min_version .'+', E_USER_ERROR);
	}

	if ($aseco->server->getGame() != 'TMF') {
		trigger_error('[plugin.ultimania.php] This plugin supports only TMF, can not start with a "'. $aseco->server->getGame() .'" Dedicated-Server!', E_USER_ERROR);
	}

	// Read Configuration
	$cfg = simplexml_load_file('ultimania.xml');
	
	/*
	$ulti_cfg['messages']['record_new']			= $cfg->record_new;
	$ulti_cfg['messages']['record_equal']		= $cfg->record_equal;
	$ulti_cfg['messages']['record_new_rank']	= $cfg->record_new_rank;
	$ulti_cfg['messages']['record_first']		= $cfg->record_first;
	*/
	
	$ulti_cfg['api_url']				= trim(file_get_contents(ULTI_API_INFO.'url.txt')).'TMF/'.ULTI_API_VERSION.'/index.php';
	$aseco->console('[plugin.ultimania.php] The API URL have been set to '.$ulti_cfg['api_url']);
	if (!$ulti_cfg['api_url']) trigger_error('[plugin.ultimania.php] Unable to get API URL from '.$ulti_cfg['api_url'], E_USER_WARNING);
	
	$ulti_cfg['window_welcome'] 		= file_get_contents(ULTI_API_INFO.'tmf/description_window.txt');

	$ulti_cfg['refresh_interval']		= $cfg->refresh_interval;
	if ($ulti_cfg['refresh_interval'] < 5) trigger_error('[plugin.ultimania.php] Refresh interval too low! Please check the config!', E_USER_ERROR);
	
	$ulti_cfg['connect_timeout']		= $cfg->connect_timeout;
	
	$ulti_cfg['request_timeout']		= $cfg->request_timeout;
	
	$ulti_cfg['retry']					= (time() + 3);
	
	$ulti_cfg['debug']					= ((strtoupper($cfg->debug) == 'TRUE') ? true : false);
	
	$ulti_cfg['timeout']				= $cfg->request_timeout;
	
	$ulti_cfg['gui']					= $cfg->gui;


	// Debug?
	if ($ulti_cfg['debug'] == true) {
		// Create debugging log file
		$filename = './debug_ultimania.log';
		if (!$fh = fopen($filename, "a", 0) ) {
			trigger_error('[plugin.ultimania.php] Could not create debugging log file "'. $filename .'"!', E_USER_ERROR);
		} else {
			$ulti_cfg['debug_logfile'] = $fh;
		}
	}
	
	// when records loaded, refresh
	ultimania_widgetShow($aseco, null);
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_onShutdown($aseco) {
	global $ulti_cfg;
	
	// Debug?
	if ($ulti_cfg['debug'] == true) {
		// Close debugging log file
		fclose($ulti_cfg['debug_logfile']);
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

// called @ onNewChallenge2
function ultimania_onNewChallenge2($aseco, $challenge_item) {
	global $ulti_cfg, $ultimania_db;

	// Debug?
	if ($ulti_cfg['debug'] == true) {
		// Write a notice
		fwrite($ulti_cfg['debug_logfile'], 'ultimania_onNewChallenge2() called at Gamemode "'. $aseco->server->gameinfo->mode .'" for Challenge uid "'. $aseco->server->challenge->uid .'"' . "\n");
	}

	// Destroy the global records array
	$ultimania_db = false;
	
	/*
	// Bail out immediately if not 'Stunts'
	if ($aseco->server->gameinfo->mode != 4) {
		return;
		trigger_error('[plugin.ultimania] Unsupported GameMode: Stunts', E_USER_ERROR);
	}
	*/
	
	// Bail out without challenge UID
	if (!$aseco->server->challenge->uid ) {
		$aseco->console('[plugin.ultimania] Could not read Challenge UID!');
		return;
	}
	
	$aseco->client->query('GetServerName');
	$servername = $aseco->client->getResponse();

	// Get the current records for this challenge
	$param = array(
		'action'		=> 'gettop',
		'uid'			=> $aseco->server->challenge->uid,
		'mapname'		=> $aseco->server->challenge->name,
		'stunt'			=> ($aseco->server->gameinfo->mode == 4) ? 1 : 0,
		'server'		=> $aseco->server->serverlogin,
		'servername'	=> $servername
	);
	$recs = ultimania_sendRequest(
		$ulti_cfg['api_url'],
		$param
	);
	
	// Debug?
	if ($ulti_cfg['debug'] == true) {

		ob_start();
		var_dump($recs);
		$dump = ob_get_clean();

		// Write a notice
		fwrite($ulti_cfg['debug_logfile'], 'ultimania_sendRequest() results (dump): '. $dump . "\n");
	}
	
	// Did this challenge have saved ultimania records?
	if (isset($recs[0])) {
		// yes - theres an existing array index 0
		// copy db into the global var
		$ultimania_db = $recs;
	} else {
		// Setup empty array
		$ultimania_db = array();
	}

	// Release 'loaded' event
	$aseco->releaseEvent('onUltimaniaRecordsLoaded', $ultimania_db);

	// Debug?
	if ($ulti_cfg['debug'] == true) {

		ob_start();
		var_dump($ultimania_db);
		$dump = ob_get_clean();

		// Write a notice
		fwrite($ulti_cfg['debug_logfile'], 'ultimania_onNewChallenge2() $ultimania_db (dump) after work: '. $dump . "\n########################################\n");
	}

}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/
function ultimania_onPlayerFinish($aseco, $finish_item) {
	global $ulti_cfg, $ultimania_db;
	
	$login = $finish_item->player->login;

	// If no actual finish, bail out immediately
	if ($finish_item->score == 0) return;
	
	// search for the Player in the records list
	/*
	foreach ($ultimania_db as $entry) {
		if ($entry['login'] == $login) {
			if ($aseco->server->gameinfo->mode == 4) {
				if ($finish_item->score <= $entry['score']) return;
			} else {
				if ($finish_item->score >= $entry['score']) return;
			}
		}
	}
	// player wasn't found, so we can upload the record
	*/

	/*
	// Get current Gamemode
	$gamemode = $aseco->server->gameinfo->mode;

	// Check if the current GameMode is 'Stunts', do nothing if not!
	if ($gamemode != 4) return;
	*/
	
	// Remove sizes
	$nickname = htmlspecialchars($finish_item->player->nickname);
	
	$aseco->client->query('GetServerName');
	$servername = $aseco->client->getResponse();
	
	// Send the new record
	$param = array(
		'action'		=> 'playerfinish',
		'uid'			=> $finish_item->challenge->uid,
		'login'			=> $login,
		'nick'			=> $nickname,
		'score'			=> intval($finish_item->score),
		'stunt'			=> ($aseco->server->gameinfo->mode == 4) ? 1 : 0,
		'server'		=> $aseco->server->serverlogin,
		'servername'	=> $servername
	);
	$response = ultimania_sendRequest(
		$ulti_cfg['api_url'],
		$param
	);
	
	if ($response['banned'] == true) {
		$message = ('$f00Your Records wont be saved on Ultimania because you\'re banned (Reason: '.$response['reason'].')');
		$aseco->client->query('ChatSendServerMessageToLogin', $message, $login);
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_onPlayerConnect($aseco, $player) {
	global $ulti_cfg;
	
	$banned_players = ultimania_getBannedPlayers();
	
	// No players banned :)
	if ($banned_players == 'null') return;
	
	$match = false;
	foreach ($banned_players as $id => $data) {
		if ($data['login'] == $player->login) {
			$match = $data;
			
			// display information window for the banned player
			$header = 'Ultimania global record database information:';
			$data = array();
			$data[] = array('$f00You\'re banned from the global records database Ultimania!');
			$data[] = array('');
			$data[] = array('This means:');
			$data[] = array('- You can\'t drive records anymore');
			$data[] = array('- On every join on a server, a message like "XY is banned from Ultimania" will appear');
			$data[] = array('- You always see this window on server join');
			$data[] = array('- Everyone playing with you can see, you\'re banned as long, as you\'re banned');
			$data[] = array('');
			$data[] = array('You can be unbanned by writing a formal apology for whatever you done');
			$data[] = array('(cheating, hacking, ...) to enwi2@t-online.de (the mail of the owner of Ultimania, Askuri).');
			$data[] = array('If we from Ultimania think, you regret whatever you done, we\'ll unban you.');
			$data[] = array('If not, we\'ll answer your mail but not unban you. In this case we can just say:');
			$data[] = array('Whatever you done, you hopefully don\'t do it again.');
			$data[] = array('');
			$data[] = array('Best regards,');
			$data[] = array('The whole Ultimania team');
			display_manialink($player->login, $header, array('Icons64x64_1', 'TrackInfo', -0.01), $data, array(0.9), 'OK');
		}
	}
	
	if ($match !== false) {
		$message = ('$f00The player '.ultimania_handleSpecialChars($player->nickname).'$z$f00 ($fff'.$player->login.'$f00) is banned from Ultimania! Reason: '.$match['reason']);
		$aseco->client->query('ChatSendServerMessage', $message);
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_getBannedPlayers() {
	global $aseco, $ulti_cfg;
	
	$aseco->client->query('GetServerName');
	$servername = $aseco->client->getResponse();
	
	$param = array(
		'action'		=> 'getbannedplayers',
		'server'		=> $aseco->server->serverlogin,
		'servername'	=> $servername
	);
	
	return ultimania_sendRequest(
		$ulti_cfg['api_url'],
		$param
	);
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_sendRequest($url, $params = array()) {
	global $aseco, $ulti_cfg;
	
	if (!function_exists('curl_init')){
		trigger_error('[plugin.ultimania.php] You need to enable cURL in PHP', E_USER_ERROR);
	}
	
	// $aseco->console('[plugin.ultimania] Sending request to API ...');
	// $aseco->console('[plugin.ultimania] URL: '.$url);
	
	$ch = curl_init();
	
	curl_setopt($ch, CURLOPT_URL, 				$url);
	curl_setopt($ch, CURLOPT_USERAGENT, 		'TMF\Xaseco'.XASECO_VERSION.'\Ultimania'.ULTI_VERSION.'\API'.ULTI_API_VERSION);
	curl_setopt($ch, CURLOPT_HEADER, 			false);
	curl_setopt($ch, CURLOPT_RETURNTRANSFER, 	true);
	curl_setopt($ch, CURLOPT_POST, 				true);
	curl_setopt($ch, CURLOPT_POSTFIELDS, 		$params);
	curl_setopt($ch, CURLOPT_TIMEOUT, 			$ulti_cfg['connect_timeout']);
	curl_setopt($ch, CURLOPT_LOW_SPEED_TIME, 	$ulti_cfg['request_timeout']);
	
	$raw = curl_exec($ch);
	curl_close($ch);
	
	// $raw = str_replace('\\', '\\\\', $raw);
	// $raw = utf8_encode($raw);
	
	// if ($raw == 'null') return;
	
	$result = json_decode($raw, true);
	
	if (($result === null) and (!empty($raw))) {
		trigger_error('[plugin.ultimania.php] Could not json_decode API response! Please report this to enwi2@t-online.de! Raw result:', E_USER_WARNING);
		var_dump($raw).CRLF;
	}
	
	// print_r($result).PHP_EOL;
	
	if (!empty($result['error']) ) {
		$aseco->console('[plugin.ultimania.php] Error occurred on action '.$params['action'].': '. $result['error']);

		// Set retry
		$ulti_cfg['retry'] = (time() + $ulti_cfg['refresh_interval']);

		// Debug?
		if ($ulti_cfg['debug'] == true) {
			fwrite($ulti_cfg['debug_logfile'], 'ultimania_sendRequest() Error occurred: '. $result['error'] . "\n########################################\n");
		}
	}
	
	return $result;
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_getUID($id) {
	$query = 'SELECT Uid FROM challenges
			  WHERE Id="'.$id.'"';
	$res = mysql_query($query);
	
	if (mysql_num_rows($res) > 0) {
		$row = mysql_fetch_row($res);
		$rtn = $row[0];
	} else {
		$rtn = 0;
	}
	
	mysql_free_result($res);
	
	return $rtn;
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_getPlayerLogin($id) {
	$query = 'SELECT Login FROM players
			  WHERE Id="'.$id.'"';
	$res = mysql_query($query);
	
	if (mysql_num_rows($res) > 0) {
		$row = mysql_fetch_row($res);
		$rtn = $row[0];
	} else {
		$rtn = 0;
	}
	
	mysql_free_result($result);
	
	return $rtn;
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/
// stolen from records_eyepiece
function ultimania_handleSpecialChars ($string) {
	// Remove links, e.g. "$(L|H|P)[...]...$(L|H|P)"
	$string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)\$(L|H|P)/i', '$2', $string);
	$string = preg_replace('/\${1}(L|H|P)\[.*?\](.*?)/i', '$2', $string);
	$string = preg_replace('/\${1}(L|H|P)(.*?)/i', '$2', $string);

	// Remove $S (shadow)
	// Remove $H (manialink)
	// Remove $W (wide)
	// Remove $I (italic)
	// Remove $L (link)
	// Remove $O (bold)
	// Remove $N (narrow)
	$string = preg_replace('/\${1}[SHWILON]/i', '', $string);
	
	// Convert &
	// Convert "
	// Convert '
	// Convert >
	// Convert <
	$string = str_replace(
			array(
				'&',
				'"',
				"'",
				'>',
				'<'
			),
			array(
				'&amp;',
				'&quot;',
				'&apos;',
				'&gt;',
				'&lt;'
			),
			$string
	);
	$string = stripNewlines($string);	// stripNewlines() from basic.inc.php

	return validateUTF8String($string);	// validateUTF8String() from basic.inc.php
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_widgetShow($aseco, $records) {
	global $ultimania_db, $ulti_cfg;
	
	$gui_cfg = $ulti_cfg['gui']->widget;
	
	if ($gui_cfg->show == 'false') return;
	
	foreach ($aseco->server->players->player_list as $curr_player) {
		$xml = '
			<manialink version="1" id="ulitmania_widget"><frame posn="'.$gui_cfg->posn.'">
			
			<quad posn="0 0 0" sizen="16 38.5" style="'.$gui_cfg->background->style.'" substyle="'.$gui_cfg->background->substyle.'" action="5450101" />
			<quad posn="'.$gui_cfg->header->posn.'" sizen="'.$gui_cfg->header->sizen.'" style="'.$gui_cfg->header->style.'" substyle="'.$gui_cfg->header->substyle.'" />
			<quad posn="13.5 0.5 2" sizen="3 3" style="BgRaceScore2" substyle="LadderRank" />
			<label posn="'.$gui_cfg->header->label->posn.'" text="'.$gui_cfg->header->label->text.'" textsize="1" />
		';
		
		if (!isset($ultimania_db[0])) {
			$xml.= '<label posn="2 -5 1" text="$oCurrently'.CRLF.'no records :(" textsize="2" />';
		} else {
			$y = -3;
			$recs_count = count($ultimania_db);
			foreach ($ultimania_db as $rank => $data) {
				$rank++;
				
				// if rank is between 20 and the count of records, skip displaying
				if (($rank >= 20) and ($rank <= ($recs_count - 1))) continue;
				
				($aseco->server->gameinfo->mode == 4) ? $score = $data['score'] : $score = ultimania_formatTime($data['score']);
				
				/*
				$type = $aseco->server->challenge->gbx->type;
				
				if ($type == 'Stunts') {
					$score = $data['score'];
				} else {
					$score = ultimania_formatTime($data['score']);
				}
				*/
				
				$textstyle = ($data['login'] == $curr_player->login) ? 'TextTitle2Blink' : 'TextTitle2';
				
				switch ($rank) {
					case 1:
						$xml.= '<quad posn="1.5 '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="First" />';
						break;
					case 2:
						$xml.= '<quad posn="1.5 '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Second" />';
						break;
					case 3:
						$xml.= '<quad posn="1.5 '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Third" />';
						break;
					
					default:
						$xml.= '<label posn="1.2 '.$y.' 1" text="'.$rank.'." textsize="1" />';
				}
				
				$xml.= '<label posn="3.5 '.$y.' 1" text="'.$score.'" textsize="1" />';
				$xml.= '<label posn="8.5 '.$y.' 1" text="'.ultimania_handleSpecialChars($data['nick']).'" style="'.$textstyle.'" sizen="13.5 2.5" textsize="1" scale="0.5" />';
				$y -= 1.75;
			}
		}
		
		$xml.= '</frame></manialink>';
			
		$aseco->client->addCall('SendDisplayManialinkPageToLogin', array($curr_player->login, $xml, 0, false));
		// $aseco->client->addCall('SendDisplayManialinkPage', array($xml, 0, false));
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_widgetHide($aseco) {
	$xml = '<manialink version="1" id="ulitmania_widget"></manialink>';
		
	$aseco->client->addCall('SendDisplayManialinkPage', array($xml, 0, false));	
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_windowShow($aseco, $login) {
	global $ultimania_db, $ulti_cfg;
	
	$xml = '<manialink version="1" id="ulitmania_window"> <frame scale="'.$scale.'">
		
		<quad posn="0 3 0" sizen="79 54" bgcolor="001a" halign="center" valign="center" />
		<quad posn="0 3 2" sizen="81 56" style="Bgs1InRace" substyle="BgCard3" halign="center" valign="center" />
		
		<quad posn="0 29.5 1" sizen="82 5" style="BgsPlayerCard" substyle="BgRacePlayerLine" halign="center" />
		<quad posn="0 30 0.5" sizen="80 5" style="Bgs1InRace" substyle="BgTitle3_3" halign="center" />
		<label posn="0 29 2" style="TextRankingsBig" halign="center" text="Ultimania World Records" />
		
		<quad posn="33.5 30 3" sizen="5 5" style="Icons128x128_1" substyle="BackFocusable" action="5450102" />
		
		<label posn="-38 -23.65 3" text="$000$tUltimania TMF v'.ULTI_VERSION.' using API '.ULTI_API_VERSION.'" textsize="0" />
	';
	
	$ultinfo = $ulti_cfg['window_welcome'];
	
	$xml.= '
		<quad posn="-38 1 1" sizen="17.5 47" style="BgsPlayerCard" substyle="BgRacePlayerName" valign="center" />
		<quad posn="-20 1 1" sizen="17.5 47" style="BgsPlayerCard" substyle="BgRacePlayerName" valign="center" />
		<quad posn="-2 24.5 1" sizen="40 20" style="BgsPlayerCard" substyle="BgRacePlayerName" />
		<quad posn="-2 -14.5 1" sizen="40 8" style="BgsPlayerCard" substyle="BgRacePlayerName" />
		
		<!-- DO	NOT REMOVE THESE BANNERS!!! THE HOSTING OF ULTIMANIA IS PAYED WITH HELP OF THESE PARTNER-LINKS! -->
		<!-- So please, please keep them as they are, thanks! -->
		<quad posn="-2 4 3" sizen="44 20" image="http://www.netcup.de/bilder/banner/netcup-hlogo-b440h200.png" scale="0.9" />
		<quad posn="-2 4 2" sizen="40 18" style="BgsPlayerCard" substyle="BgRacePlayerName" />
		<quad posn="-2 4 1" sizen="40 18" style="BgsPlayerCard" substyle="ProgressBar" url="http://askuri.de/nc_gutscheine.php" />
		<!-- BANNERS ENDING HERE -->
		
		<label posn="-1.5 24 2" text="'.$ultinfo.'" textsize="2" />
		
		<label posn="5 -16.5 2" text="Cheated record? Report this Map!" style="CardButtonMedium" action="5450103" />
	';
	
	$xml.= '<frame posn="-38.5 27 1">';
	
	if (!isset($ultimania_db[0])) {
		$xml.= '<label posn="2 -5 1" text="$oCurrently'.CRLF.'no records :(" textsize="2" />';
	} else {
		$x = 0;
		$y = -3;
		$recs_count = count($ultimania_db);
		foreach ($ultimania_db as $rank => $data) {
			$rank++;
			
			$type = $aseco->server->gameinfo->mode;
			
			if ($type == 4) {
				$score = $data['score'];
			} else {
				$score = ultimania_formatTime($data['score']);
			}
			
			switch ($rank) {
				case 1:
					$xml.= '<quad posn="'.($x + 1.5).' '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="First" />';
					break;
				case 2:
					$xml.= '<quad posn="'.($x + 1.5).' '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Second" />';
					break;
				case 3:
					$xml.= '<quad posn="'.($x + 1.5).' '.$y.' 1" sizen="1.5 1.5" style="Icons64x64_1" substyle="Third" />';
					break;
				
				default:
					$xml.= '<label posn="'.($x + 1.2).' '.$y.' 1" text="'.$rank.'." textsize="1" />';
			}
			
			$xml.= '<label posn="'.($x + 3.5).' '.$y.' 1" text="'.$score.'" textsize="1" />';
			$xml.= '<label posn="'.($x + 8.5).' '.$y.' 1" text="'.ultimania_handleSpecialChars($data['nick']).'" sizen="7.5 2" textsize="1" />';
			
			if ($data['login'] == $login) {
				$xml.= '<quad posn="'.($x + 0.7).' '.($y + 0.1).' 0.5" sizen="17 2" style="BgsPlayerCard" substyle="ProgressBar" />';
			}
			
			// $xml.= '<quad posn="'.($x + 16.3).' '.($y - 0.1).' 2" sizen="1.3 1.3" style="Icons64x64_1" substyle="TrackInfo" action="54501030'.$rank.'" />';
			// $xml.= '<quad posn="'.($x + 16).' '.($y).' 1" sizen="2 2" style="BgsPlayerCard" substyle="ProgressBar" />';
			
			$y -= 1.75;
			
			if ($rank == 25) {
				$x+= 18;
				$y = -3;
			}
		}
	}
	
	$xml.= '</frame> </frame> </manialink>';
		
	$aseco->client->addCall('SendDisplayManialinkPageToLogin', array($login, $xml, 0, false));
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_windowHide($aseco, $login) {
	$xml = '<manialink version="1" id="ulitmania_window"></manialink>';
	
	if ($login === false) {
		$aseco->client->addCall('SendDisplayManialinkPage', array($xml, 0, false));	
	} else {
		$aseco->client->addCall('SendDisplayManialinkPageToLogin', array($login, $xml, 0, false));
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

// called @ onEverySecond
function ultimania_onEverySecond($aseco) {
	global $ulti_cfg;
	
	// execute only, if status is 4: running - play
	if ($aseco->currstatus != 4) return;
	
	if (($ulti_cfg['retry'] != 0) and (time() >= $ulti_cfg['retry'])) {
		$ulti_cfg['retry'] = time() + $ulti_cfg['refresh_interval'];
		
		$players_count = count($aseco->server->players->player_list);
		if ($players_count >= 1) {
			// refresh db
			ultimania_onNewChallenge2($aseco, null);
			
			// when records loaded, refresh
			ultimania_widgetShow($aseco, null);
		}
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

// Stolen from plugin.records_eyepiece.php / undef, who stole it from basic.inc.php and adjusted it
function ultimania_formatTime ($MwTime, $hsec = true) {

	if ($MwTime == -1) {
		return '???';
	}
	else {
		$hseconds = (($MwTime - (floor($MwTime/1000) * 1000)) / 10);
		$MwTime = floor($MwTime / 1000);
		$hours = floor($MwTime / 3600);
		$MwTime = $MwTime - ($hours * 3600);
		$minutes = floor($MwTime / 60);
		$MwTime = $MwTime - ($minutes * 60);
		$seconds = floor($MwTime);
		if ($hsec) {
			if ($hours) {
				return sprintf('%d:%02d:%02d.%02d', $hours, $minutes, $seconds, $hseconds);
			}
			else {
				return sprintf('%d:%02d.%02d', $minutes, $seconds, $hseconds);
			}
		}
		else {
			if ($hours) {
				return sprintf('%d:%02d:%02d', $hours, $minutes, $seconds);
			}
			else {
				return sprintf('%d:%02d', $minutes, $seconds);
			}
		}
	}
}

/*
#///////////////////////////////////////////////////////////////////////#
#									#
#///////////////////////////////////////////////////////////////////////#
*/

function ultimania_reportMap ($aseco, $login) {
	global $ulti_cfg;
	
	$aseco->client->query('GetServerName');
	$servername = $aseco->client->getResponse();
	
	$param = array(
		'action'		=> 'reportmap',
		'uid'			=> $aseco->server->challenge->uid,
		'stunt'			=> ($aseco->server->gameinfo->mode == 4) ? 1 : 0,
		'server'		=> $aseco->server->serverlogin,
		'servername'	=> $servername
	);
	$response = ultimania_sendRequest(
		$ulti_cfg['api_url'],
		$param
	);
	
	$aseco->client->query('ChatSendServerMessageToLogin', '$0f0Thank you for supporting Ultimania!', $login);
	$aseco->client->query('ChatSendServerMessageToLogin', '$0f0A Ultimania Mod will look at the map in the next time.', $login);
	
	ultimania_windowHide($aseco, $login);
	
}

?>
